/*

 usb.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <errno.h>

#include "usb.h"
#include "crc8.h"

#ifndef WIN32
#include <ftdi.h>

static struct ftdi_context ftDevice;
#else
#include <windows.h>	//also in ideservd.h, but we need it now
#include "ftd2xx.h"

FT_STATUS ftStatus;
FT_HANDLE ftHandle;
#endif //WIN32

static char *drivername(void) {
    return "USB";
}

static int initialize(char *dev, int unused) {
#ifndef WIN32
    ftdi_init(&ftDevice);
    if (ftdi_usb_open(&ftDevice, 0x0403, 0x6001)) {
	fprintf(stderr, "ftdi_usb_open: %s\n", ftDevice.error_str);
        return -1;
    }

    if (ftdi_usb_reset(&ftDevice)) {
        fprintf(stderr, "ftdi_usb_reset: %s\n", ftDevice.error_str);
        return -1;
    }
    ftdi_usb_purge_buffers(&ftDevice);
    ftdi_set_latency_timer(&ftDevice, 8);
#else
    ftStatus = FT_Open(0, &ftHandle);
    if(ftStatus != FT_OK) {
		/* 
			This can fail if the ftdi_sio driver is loaded
		 	use lsmod to check this and rmmod ftdi_sio to remove
			also rmmod usbserial
		 */
		fprintf(stderr,"FT_Open failed\n");
		exit(1);
    }

    FT_ResetDevice(ftHandle);
    FT_Purge(ftHandle, FT_PURGE_RX|FT_PURGE_TX);
    FT_SetTimeouts(ftHandle, 100, 10000);	// C64->PC, PC->C64
    FT_SetLatencyTimer(ftHandle, 5);
//     FT_SetUSBParameters(ftHandle, 64,0);
//     FT_SetDeadmanTimeout(ftHandle, 10);
#endif
    return 0;
}

static int getb(void) {
    unsigned char data;
    int err;

#ifndef WIN32
    err=ftdi_read_data(&ftDevice, &data, 1);
    if (err == -ENODEV) return -ENODEV;
    if (err < 0) return -EIO;
    if (err == 1) {
        crc_add(data);
        return (unsigned char)data;
    }
#else
    DWORD dwBytesRead;

    err=FT_Read(ftHandle, &data, 1, &dwBytesRead);
    if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
    if (err != FT_OK) return -ENODEV;
    if (dwBytesRead == 1) {
        crc_add(data);
        return (unsigned char)data;
    }
#endif
    return -EBUSY;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    int err;
#ifdef WIN32
    DWORD dwBytesRead;
#endif

    do {
	int l;
	if (bytes>4096) l=4096; else l=bytes;
#ifndef WIN32
        err=ftdi_read_data(&ftDevice, data, l);
        if (err == -ENODEV) return -ENODEV;
        if (err < 0 ) return -EIO;
	for (l=0;l < err;l++) crc_add(data[l]);
	bytes-=err;data+=err;
#else
        err=FT_Read(ftHandle, data, l, &dwBytesRead);
        if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
        if (err != FT_OK) return -EIO;
        for (l=0;l < dwBytesRead;l++) crc_add(data[l]);
	bytes-=dwBytesRead;data+=dwBytesRead;
#endif
    } while (bytes > 0);
    return 0;
}

static int sendb(unsigned char data) {
    int err;
    crc_add(data);
#ifndef WIN32
    do {
        err=ftdi_write_data(&ftDevice, &data, 1);
        if (err == -ENODEV) return -ENODEV;
        if (err < 0 ) return -EIO;
    } while (err != 1);
#else
    DWORD dwBytesWritten;
    do {
        err=FT_Write(ftHandle, &data, 1, &dwBytesWritten);
        if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
        if (err != FT_OK) return -EIO;
    } while (dwBytesWritten != 1);
#endif
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    int err;
#ifdef WIN32
    DWORD dwBytesWritten;
#endif

    for (err=0;err < bytes;err++) crc_add(data[err]);
    do {
	int l;
	if (bytes > 4096) l=4096; else l=bytes;
#ifndef WIN32
	err=ftdi_write_data(&ftDevice, data, l);
        if (err == -ENODEV) return -ENODEV;
        if (err < 0 ) return -EIO;
	bytes-=err;data+=err;
#else
	err=FT_Write(ftHandle, data, l, &dwBytesWritten);
        if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
        if (err != FT_OK) return -EIO;
	bytes-=dwBytesWritten;data+=dwBytesWritten;
#endif
    } while (bytes > 0);
    return 0;
}	

static void shutdown(void) {
#ifndef WIN32
    ftdi_usb_purge_buffers(&ftDevice);
    ftdi_usb_close(&ftDevice);
    ftdi_deinit(&ftDevice);
#else
    FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
    FT_Close(ftHandle);
#endif
}

static int flush(void) {
    return 0;
}

static void clean(void) {
#ifndef WIN32
    ftdi_usb_purge_buffers(&ftDevice);
#else
    FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
#endif
}

static int wait(void) {
    int dyntime=1;
    int b;
    for (;;) {
	b=getb();
        if (b != -EBUSY) return b;
        usleep(dyntime);
        if (dyntime < 16384) dyntime<<=1;
    }
}

const struct io_ops usb_io_ops = {
    .drivername = drivername,
    .initialize = initialize,
    .getb       = getb,
    .getbytes   = getbytes,
    .sendb      = sendb,
    .sendbytes  = sendbytes,
    .shutdown   = shutdown,
    .flush      = flush,
    .wait       = wait,
    .clean      = clean,
};
